/*
 * Copyright (C) 1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <string.h>	/* for strcmp() */

#include "monitor.h"	/* for monitor{Enter,Exit}() */
#include "threads.h"	/* for JavaStackSize */
#include "compiler.h"


/*
 * Compile and invoke the method.
 */
bool_t compileAndInvokeMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee) {
  int key;

  int access;
  bool_t (*norm_invoker)(JHandle*,struct methodblock *,int,ExecEnv*);

#ifdef COMPILE_DEBUG
  printf("\n");
  printf("compileAndInvokeMethod() is called:\n  %s#%s %s\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
  fflush(stdout);
#endif

  /* if compiler is disabled, call original invoker */
  if (!compiler_enabled) {
#ifdef RUNTIME_DEBUG
    printf("Compiler is disabled, now call original invoker.");
    fflush(stdout);
#endif
    access = mb->fb.access;
    norm_invoker = access2invoker(access);
    goto candi_call_normal_method;
  }

  key = obj_monitor(fieldclass(&mb->fb));
#ifdef COMPILE_DEBUG
  printf("  key: %d\n", key);
  fflush(stdout);
#endif

  monitorEnter(key);

  /* compile the called method */
#if 0	/* this check is not needed. */
  if (!(mb->CompiledCode) &&
	!(mb->fb.access & (ACC_ABSTRACT | ACC_NATIVE)))
#else
  if (!(mb->CompiledCode))
#endif	/* if 0 */
  {
#ifdef COMPILE_DEBUG
    printf("  now compiling.\n");  fflush(stdout);
#endif

    ee->current_frame->optop += args_size;	/* to save real args */
	/* not to be broken args by method invocation during compilation */

    if (!compileMethod(mb)) {
      /* success */
      ee->current_frame->optop -= args_size;	/* restore real optop */
    }
    else {
      /* fail to compile... */
      ee->current_frame->optop -= args_size;	/* restore real optop */
#ifdef COMPILE_DEBUG
      printf("  compilation failed...\n");  fflush(stdout);
#endif
      access = mb->fb.access;
      access &= ~ACC_MACHINE_COMPILED;
      mb->fb.access = access;

#if 0	/* can't restack if caller of this method is compiled method... */
      norm_invoker = mb->invoker =
	(bool_t (*)(JHandle*,struct methodblock *,int,ExecEnv*))
		access2invoker(access);

      goto candi_call_normal_method;
#else
      return FALSE;
#endif	/* if 0 */
    }
  }

#ifdef COMPILE_DEBUG
  printf("compileAndInvokeMethod(): now call invoker.\n");
  fflush(stdout);
#endif
  monitorExit(key);
  return mb->invoker(o, mb, args_size, ee);

candi_call_normal_method:
  {
    bool_t invoker_ret;

    invoker_ret = norm_invoker(o, mb, args_size, ee);
    if (!invoker_ret)  return FALSE;

    if (!(access & (ACC_ABSTRACT | ACC_NATIVE))) {
	/* normal Java method */

      /* restack is done in invokeMethod() (in runtime.c)
	 if this method is called by compiled method */

      /* call ExecuteJava() */
      {
	int exec_ret;
	stack_item *optop, *old_optop;
	int retsize;
	JavaFrame *cur_frame;;

	cur_frame = ee->current_frame;
	exec_ret = ExecuteJava(mb->code, ee);
	old_optop = cur_frame->optop;
#ifdef EXECUTEJAVA_IN_ASM
	if (cur_frame->monitor)  monitorExit(obj_monitor(o));
#endif	/* EXECUTEJAVA_IN_ASM */
	cur_frame = cur_frame->prev;
	ee->current_frame = cur_frame;
	if (!exec_ret)  return FALSE;

#ifdef EXECUTEJAVA_IN_ASM
      /* These operation is required
	 only with x86 assembly ver. of executeJava.c */
	retsize = ((CodeInfo *)mb->CompiledCodeInfo)->ret_size;
	if (retsize != 0) {
	  optop = cur_frame->optop;
	  if (retsize == 1) {
	    optop[0] = old_optop[-1];  optop++;
	  }
	  else {	/* retsize == 2 */
	    optop[0] = old_optop[-2];  optop[1] = old_optop[-1];  optop += 2;
	  }
	  cur_frame->optop = optop;
	}
#endif	/* EXECUTEJAVA_IN_ASM */
      }	/* call ExecuteJava() */
    }	/* normal Java method */

    return invoker_ret;
  }
}


/*
 * Invoke the compiled method.
 */
bool_t invokeJITCompiledMethod(
	JHandle *o, struct methodblock *mb, int args_size, ExecEnv *ee) {
  JavaFrame *old_frame, *frame;
  struct methodblock *prev_method;	/* for restack */
  int access;
  stack_item *native_vars;
#ifdef RUNTIME_DEBUG
  int runtime_debug;
#endif

  /* allocate local var. space */
	/* caution: This code segment must be placed at code head. */
  __asm__("leal  -4(%%esp),%%edx\n\t"
	  "subl  %1,%%esp\n\t"
	  "movl  %%edx,%0"
	  : "=m" (native_vars) : "r" (mb->nlocals * 4)
	  : "edx");
	/* native_vars = %esp - 4 */
	/* %esp -= mb->nlocals * 4 */

#ifdef RUNTIME_DEBUG
#  if 0
  runtime_debug = 1;
#  else
  if ((!strcmp(cbName(mb->fb.clazz), "java/util/ResourceBundle"))
	&& (!strcmp(mb->fb.name, "findBundle")))
    runtime_debug = 1;
  else if ((!strcmp(cbName(mb->fb.clazz), "javax/swing/UIManager"))
	&& (!strcmp(mb->fb.name, "maybeInitialize")))
    runtime_debug = 1;
  else if ((!strcmp(cbName(mb->fb.clazz), "sun/tools/tree/ArrayExpression"))
	&& (!strcmp(mb->fb.name, "codeValue")))
    runtime_debug = 1;
  else if ((!strcmp(cbName(mb->fb.clazz), "java/lang/Runtime"))
	&& (!strcmp(mb->fb.name, "loadLibrary")))
    runtime_debug = 1;
  else
    runtime_debug = 0;
#  endif
#endif	/* RUNTIME_DEBUG */

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("\n");
    printf("invokeJITCompiledMethod() is called:\n  %s#%s %s\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
    if (exceptionOccurred(ee)) {
      printf("  An exception not handled remains!\n");
    }
    fflush(stdout);

#if 1
    showStackFrames(ee);
#endif
  }	/* if (runtime_debug) */
#endif	/* RUNTIME_DEBUG */

  old_frame = ee->current_frame;
  access = mb->fb.access;

  /* creates a new frame */
#if 1
  {
    JavaStack *stack = old_frame->javastack;

    frame = (JavaFrame *)(old_frame->optop + mb->nlocals);
    if (frame->ostack + mb->maxstack >= stack->end_data) {
      if (stack->next)  stack = stack->next;
      else {
	if (stack->stack_so_far + JAVASTACK_CHUNK_SIZE * sizeof(stack_item)
		> JavaStackSize) {
	  SignalError(ee, JAVAPKG "StackOverflowError", 0);
	  return FALSE;
	}
	if (!(stack = CreateNewJavaStack(ee, stack))) {
	  SignalError(ee, JAVAPKG "OutOfMemoryError", 0);
	  return FALSE;
	}
      }
      frame = (JavaFrame *)(stack->data + mb->nlocals);
	/* needless to copy args to a frame on new stack chunk */
    }
    /* frame->constant_pool = cbConstantPool(fieldclass(&mb->fb)); */
    frame->returnpc = frame->lastpc = mb->code;
	/* lastpc is not initialized in invoke*JavaMethod() */
    frame->optop = frame->ostack;
    frame->vars = old_frame->optop;
    frame->prev = old_frame;
    frame->javastack = stack;
    frame->current_method = mb;

    ee->current_frame = frame;
  }	/* create a new frame */
#else
  if (access & ACC_SYNCHRONIZED)
    invokeSynchronizedJavaMethod(o, mb, args_size, ee);
  else
    invokeJavaMethod(o, mb, args_size, ee);

  frame = ee->current_frame;
  frame->lastpc = mb->code;
	/* lastpc is not initialized in invoke*JavaMethod() */
#endif

  /* call */

  /* restack from JVM stack to native stack */
  prev_method = old_frame->current_method;
  if (prev_method) {
    if (prev_method->CompiledCode) {
      /* called by compiled method */
      frame->vars = (stack_item *)old_frame->returnpc;
      goto restack_done;
    }
  }

  /* called by normal Java or native method */
  {	/* restack */
    char *argsizes;
    stack_item *args;

    argsizes = ((CodeInfo *)mb->CompiledCodeInfo)->argsizes;
    args = frame->vars;	/* points to JVM stack */

    frame->vars = native_vars;

    while (*argsizes) {
      if (*argsizes == 1) {
	native_vars[0] = args[0];
	native_vars--;  args++;
      }
      else {
	native_vars[-1] = args[0];
	native_vars[0] = args[1];
	native_vars -= 2;  args += 2;
      }
      argsizes++;
    }	/* while (*argsizes) */
  }	/* restack */

restack_done:

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    CodeInfo *info = (CodeInfo *)mb->CompiledCodeInfo;

    printf("call compiled method.\n");
    printf("  o:0x%08x, mb:0x%08x, args_size:%d, ee:0x%08x\n",
	(int)o, (int)mb, args_size, (int)ee);
    printf("  mb->CompiledCode: 0x%08x\n", (int)mb->CompiledCode);
    printf("  code size: 0x%08x\n", (info ? info->code_size : 0));
    fflush(stdout);

    {
      stack_item *vars = ee->current_frame->vars;
      char *sig;
      int i;

      if (args_size > 0) {
	printf("  args:\n");
	i = 0;
	if (!(access & ACC_STATIC)) {
	  printf("    0x%08x @ 0x%08x  ", (int)vars[0].h, (int)vars);
	  showObjectBody("L;", vars[0].h);
	  printf("\n");
	  i++;
	}
	sig = mb->fb.signature;
	if (sig[0] == '(')  sig++;

	for (; i < args_size; i++) {
	  JHandle *arg = vars[-i].h;
	  if ((sig[0] == 'D') || (sig[0] == 'J'))  i++;
	  printf("    0x%08x @ 0x%08x  ", (int)arg, (int)(vars - i));
	  sig = showObjectBody(sig, arg);
	  printf("\n");
	}
      }
      fflush(stdout);
    }
  }
#endif


  if (!(access & ACC_SYNCHRONIZED))
    frame->monitor = 0;
  else {
    frame->monitor = o;
    monitorEnter(obj_monitor(o));
  }

#ifdef ENABLE_PROFILING
  if (java_monitor)  frame->mon_starttime = now();
#endif

#if 0
  printf("%x\n", mb->CompiledCode);
#endif
#ifdef RUNTIME_DEBUG
  ((void (*) (JHandle *, struct methodblock *, int, ExecEnv *, int))
	mb->CompiledCode)(o, mb, args_size, ee, runtime_debug);
#else
  ((void (*) (JHandle *, struct methodblock *, int, ExecEnv *))
	mb->CompiledCode)(o, mb, args_size, ee);
#endif

  if (access & ACC_SYNCHRONIZED)	/* if (frame->monitor) */
    monitorExit(obj_monitor(frame->monitor));

#ifdef ENABLE_PROFILING
  if (java_monitor)
    java_mon(prev_method, mb, now() - frame->mon_starttime);
#endif

  ee->current_frame = frame->prev;

#ifdef RUNTIME_DEBUG
  if (runtime_debug) {
    printf("compiled method finished: %s#%s %s\n",
	cbName(mb->fb.clazz), mb->fb.name, mb->fb.signature);
    fflush(stdout);

    printf("  free memory: handle %lld obj %lld\n",
		FreeHandleMemory(), FreeObjectMemory());
    fflush(stdout);

    if (exceptionOccurred(ee)) {
      JHandle *exc = ee->exception.exc;
      printf("  exception occurred:\n");
      printf("    0x%08x(%s)\n",
		(int)exc, cbName(exc->methods->classdescriptor));
      showExcStackTrace(exc);
      fflush(stdout);
    }
    printf("\n");
    fflush(stdout);
  }	/* if (runtime_debug) */
#endif


#ifdef RUNTIME_DEBUG
  runtime_debug = 0;
#endif

  return !(exceptionOccurred(ee));
}
